"""
Detailed documentation and code examples coming soon.
"""
# External
from pandas import Series, DataFrame
import openpyxl as xl
from typing import Any
from copy import copy
import os

# Internal main
from excelbird._utils.util import (
    fill_frames,
    set_duplicate_objects_to_ref,
)
from excelbird._utils.argument_parsing import (
    combine_args_and_children_to_list,
    move_remaining_kwargs_to_dict,
)
from excelbird._utils.pass_attributes import (
    pass_dict_to_children,
    pass_attr_to_children,
)
from excelbird._utils.validation import (
    require_each_element_to_be_cls_type,
)
from excelbird.exceptions import (
    AutoOpenFileError,
    InvalidSheetName,
    ExpressionResolutionError,
)
from excelbird._layout_references import Globals
from excelbird.styles.styles import default_table_style

from excelbird.core.expression import Expr
from excelbird.core.function import Func
from excelbird.core.item import Item
from excelbird.core.gap import Gap
from excelbird.core.cell import Cell
from excelbird.core.frame import _Frame, Frame
from excelbird.core.stack import _Stack, Stack
from excelbird.core.series import _Series, Col
from excelbird.core.sheet import Sheet

from excelbird._base.container import ListIndexableById
from excelbird._base.dotdict import Style
from excelbird._base.loc import Loc


class Book(ListIndexableById):
    """
    The outer-most parent container for a layout. Only `Book` has the ability to write
    to an Excel file.

    Call ``.write(path)`` to save to an Excel file.

    * Child Type: :class:`Sheet`

    Parameters
    ----------
    *args : Union[Sheet, Stack, VStack, Frame, VFrame, Col, Row, Cell, list, tuple, str, int, float, pd.Series, pd.DataFrame, np.ndarray, Gap, Expr, Func, set]
        Each element, if not a :class:`Sheet`, will be placed in its own separate `Sheet`.
        Vectors which aren't layout types (like `list`, or `pd.DataFrame`) will be inferred
        as :class:`Col` or :class:`Frame`
    children : list, optional
        Will be combined with args
    path : str, optional
        Path to write Book. Can be omitted and passed to ``.write()`` instead
    auto_open : bool, default False
        Attempt to automatically open after calling ``.write()``. If a file with the same name
        is already open, it will be closed first. Requires dependency, xlwings
    sep : Gap or bool or int or dict, optional
        A sep in any excelbird layout element inserts a Gap between each of its children.
        If True, a default of ``Gap(1)`` is used. If int, ``Gap(sep)`` will be used. If a dict,
        ``Gap(1, **sep)`` will be used.
    tab_color : str, optional
        Applied to each child Sheet
    end_gap : bool or int or dict or Gap, optional
        Applied to each child Sheet
    isolate : bool, optional
        Applied to each child Sheet
    zoom : int, optional
        Applied to each child Sheet
    cell_style : dict, optional
        Applied to each child Sheet
    header_style : dict, optional
        Applied to each child Sheet
    table_style : dict or bool, optional
        Applied to each child Sheet
    **kwargs :
        Remaining keyword arguments applied to ``cell_style``, to be passed down to children

    """

    _dimensions = -1

    elem_type = Sheet

    def __init__(
        self,
        *args: Any,
        children: list | None = None,
        path: str | None = None,
        auto_open: bool = False,
        sep: Any | None = None,
        tab_color: str | None = None,
        end_gap: bool | int | dict | Gap | None = None,
        isolate: bool | None = None,
        zoom: int | None = None,
        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        table_style: Style | dict | bool | None = None,
        **kwargs,
    ) -> None:
        children = combine_args_and_children_to_list(args, children)

        children = [i for i in children if i is not None]

        if cell_style is None:
            cell_style = dict()
        if header_style is None:
            header_style = dict()
        if table_style is None or table_style is False:
            table_style = dict()
        elif table_style is True:
            table_style = default_table_style

        self._format_args(children)

        move_remaining_kwargs_to_dict(kwargs, cell_style)

        original_workbook = xl.Workbook()

        self.original_workbook = original_workbook
        self.wb = copy(original_workbook)
        self.path = path
        self.auto_open = auto_open
        # Attrs that must be passed to children
        self.tab_color = tab_color
        self.end_gap = end_gap
        self.isolate = isolate
        self.zoom = zoom
        # Dicts that must be passed to children
        self.cell_style = Style(**cell_style)
        self.header_style = Style(**header_style)
        self.table_style = Style(**table_style)

        self._init(children)

        if sep is not None:
            self._insert_separator(sep)

    def write(self, path: str | None = None) -> None:
        """
        Evaluates the layout tree and writes the completed layout to a ``.xlsx`` file.

        Parameters
        ----------
        path : str, optional
            Full path to the output file. Only exclude if `path` attribute
            has already been set.

        Notes
        -----

        **The algorithm, step by step**

        * First, all references inside each :class:`Expr <excelbird.Expr>` in the layout is resolved and evaluated
        * Now that the true size and shape of each layout element is known, spatial styling
          can be resolved

          * Margin and padding values are interpreted to create correctly sized :class:`Gap <excelbird.Gap>` spacing
          * Background colors are applied to the appropriate cells, drilling breadth-first through the tree.
          * Then, all `Gaps` are evaluated as correctly sized and styled layout elements.

        * Now that the layout structure, input data, and styling is complete

          * Traverse the tree and iteratevely assign a true sheet/column/row coordinate
            to each :class:`Cell <excelbird.Cell>` individually. Notice that *only* cells
            who were placed directly inside the `Book` get assigned a location. Later on,
            this lets us ensure that the formulas and cell references we create are valid.
          * Call each child's ``._write()`` method. Each layout element's write method will 
            safely pass down styling to its children, and then call each child's write method.

        * `Cells` which contain cell references don't know the value of their formula
          until their ``._write()`` is called. At this point, their expression is nothing
          more than a binary tree of references to other Python objects. When recursively expanding
          the tree to a formula string, we're able to tell whether a referenced `Cell` is actually
          being placed in the workbook or not, since unplaced `Cells` were never assigned locations.
          Therefore, we will always end up with a valid formula, because instead of trying to
          reference a non-existent cell, we take its value or formula and include it in
          our own.

        """
        if path is not None:
            self.path = path

        if self.path is None:
            raise ValueError("Workbook needs a path")

        require_each_element_to_be_cls_type(self)

        if self.auto_open == True:
            self._save_close_currently_open_excel_file()

        if self._resolve_all_references() is False:
            raise ExpressionResolutionError()

        pass_attr_to_children(self, "end_gap")

        self._validate_child_types()

        set_duplicate_objects_to_ref(self, [])

        for sheet in self:
            fill_frames(sheet)
            sheet._resolve_padding()
            sheet._resolve_margin()
            sheet._resolve_background_color()
            sheet._resolve_gaps()

        self._set_loc()

        pass_attr_to_children(self, "tab_color")
        pass_attr_to_children(self, "isolate")
        pass_attr_to_children(self, "zoom")
        pass_dict_to_children(self, "cell_style")
        pass_dict_to_children(self, "header_style")
        pass_dict_to_children(self, "table_style")

        for sheet in self:
            sheet._write()

        self.wb.save(self.path)
        print(f"Book '{self.path}' saved")
        if self.auto_open == True:
            self._open_excel_file()

        Globals.clear_references()
        Globals.clear_global_references()

    def _format_args(self, args: list) -> None:
        """
        Please refactor so that any element that isn't sheet or gap
        is simply passed to the Sheet constructor
        """

        Item._resolve_all_in_container(args, Sheet)

        elem_type = type(self).elem_type
        for i, elem in enumerate(args):
            if isinstance(elem, elem_type):
                continue

            if isinstance(elem, DataFrame):
                args[i] = elem_type(Frame(elem))

            elif isinstance(elem, Series):
                args[i] = elem_type(Col(elem))

            elif isinstance(elem, (_Series, _Frame, _Stack)):
                args[i] = elem_type(elem)

            elif isinstance(elem, Gap):
                gap = args.pop(i)
                for _ in range(gap):
                    args.insert(i, Sheet(**gap.kwargs))

            elif isinstance(elem, Cell):
                args[i] = elem_type(elem)

            elif isinstance(elem, (int, str, float)) and not isinstance(elem, Gap):
                args[i] = elem_type(Cell(elem))

            elif type(elem) in [list, tuple]:
                if len(elem) == 0:
                    args.pop(i)
                else:
                    if isinstance(elem[0], (Cell, int, str, float)):
                        args[i] = elem_type(Col(*elem))
                    elif isinstance(elem[0], (list, tuple, Series)):
                        args[i] = elem_type(Frame(*elem))
                    else:
                        args[i] = elem_type(Stack(*elem))

    def _resolve_all_references(self) -> bool:
        Expr._set_use_ref_for_container_recursive(self)

        all_resolved = False
        attempts = 0
        while not all_resolved and attempts <= 5:
            all_resolved = True
            attempts += 1
            if Expr._resolve_container_recursive(self) is False:
                all_resolved = False
            if Func._resolve_container_recursive(self) is False:
                all_resolved = False

        return all_resolved

    def _validate_child_types(self) -> None:
        valid_types = (
            _Stack,
            _Frame,
            _Series,
            Cell,
            Gap,
        )
        type_names = [e.__name__ for e in valid_types]
        for elem in self:
            if not isinstance(elem, valid_types):
                raise TypeError(
                    f"At write time, a Book can only hold the following types:\n{type_names}"
                )
            if hasattr(elem, "_validate_child_types"):
                elem._validate_child_types()

    def _set_loc(self):
        for i, sheet in enumerate(self):
            if i == 0:
                ws = self.wb.active
            else:
                ws = self.wb.create_sheet(f"Sheet{i+1}")

            if sheet.title is None:
                sheet.title = f"Sheet{i+1}"

            invalid_sheet_name_chars = [":", "\\", "/", "?", "*", "[", "]"]
            if any(c in sheet.title for c in invalid_sheet_name_chars):
                raise InvalidSheetName(
                    f"Sheet name must not contain, {invalid_sheet_name_chars}"
                )

            ws.title = sheet.title
            sheet._set_loc(Loc((0, 0), ws))

    def __repr__(self):
        return ""

    def _save_close_currently_open_excel_file(self):
        if self.auto_open is True:
            xw = self._try_to_import_xlwings()
            # Calling `xw.books` will raise error if excel is not already open
            # If excel isn't open, just call `xw.App`
            try:
                base_name = os.path.basename(self.path).lower()
                if base_name in xw.books:
                    book = xw.Book(self.path)
                    book.save()
                    book.close()

            except xw.XlwingsError as e:
                raise AutoOpenFileError(
                    "Couldn't access Excel file. A common cause of this issue is having your "
                    "file stored in OneDrive, AND it being currently open before executing your code. "
                    "To fix this, EITHER move the file out of OneDrive, OR close the file before code execution."
                    f'\nThis error was triggered by an `XlwingsError` being raised. Its message is:\n"{e}"'
                )

    def _open_excel_file(self):
        if self.auto_open is True:
            xw = self._try_to_import_xlwings()

            try:
                print("Opening Book...")
                xw.Book(self.path)

            except xw.XlwingsError as e:
                raise AutoOpenFileError(
                    "Couldn't access Excel file. A common cause of this issue is having your "
                    "file stored in OneDrive, AND it being currently open before executing your code. "
                    "To fix this, EITHER move the file out of OneDrive, OR close the file before code execution."
                    f'\nThis error was triggered by an `XlwingsError` being raised. Its message is:\n"{e}"'
                )

    def _try_to_import_xlwings(self) -> None:
        try:
            import xlwings as xw

            return xw
        except Exception:
            raise ModuleNotFoundError(
                "The `auto_open` option uses the `xlwings` library to handle opening/closing "
                "Excel sessions. Please 'pip install xlwings' to continue, or set `auto_open=False`"
            )



from excelbird.core.cell import Cell

class _Merged:
    """
    Take a Cell, and distance across/down to merge.
    Parent container will figure out how to interpret this.

    If inside a Col or Row, there is no need to differentiate between across/down.
    Just pass a single integer to `amount` as a positional.

    If inside a Col, for instance, and amount is 2, the parent will explode this value into
    three Cells: `*`[Cell(..., merge=(2,0)), Cell(), Cell()] where the first element is the initial
    Cell that was passed, but with the merge attribute set to merge 2 below, plus two empty cells
    to soak up the merge.

    If inside a Frame, and across is 2, down is 3, the parent will explode this value into 3 Cols:
        `*`[
            Col(Cell(..., merge=(3,2)), Cell(), Cell(), Cell()),
            Col(Cell(), Cell(), Cell(), Cell()),
            Col(Cell(), Cell(), Cell(), Cell())
        ]

    If inside a Stack, there is no need to explode. Insert a Frame instead:
        Frame(
            Col(Cell(..., merge=(3,2)), Cell(), Cell(), Cell()),
            Col(Cell(), Cell(), Cell(), Cell()),
            Col(Cell(), Cell(), Cell(), Cell())
        )

    """
def __init__(self, cell: Cell, across: int = None, down: int = None):
    pass



from typing import Any, Iterable


class HasMargin:
    empty = [None, None, None, None]

    def _init_margin(self, margin, top, right, bottom, left) -> None:

        if (margin, top, right, bottom, left) == (None, None, None, None, None):
            self.margin_top = None
            self.margin_right = None
            self.margin_bottom = None
            self.margin_left = None
            return

        if not isinstance(margin, list):
            margin = [margin]
        if len(margin) == 1:
            margin *= 4
        elif len(margin) == 2:
            margin *= 2
        elif len(margin) == 3:
            margin += [None]

        self.margin_top = margin[0]
        self.margin_right = margin[1]
        self.margin_bottom = margin[2]
        self.margin_left = margin[3]

        if top is not None:
            self.margin_top = top
        if right is not None:
            self.margin_right = right
        if bottom is not None:
            self.margin_bottom = bottom
        if left is not None:
            self.margin_left = left

    @property
    def margin(self) -> list[int | None]:
        return [
            self.margin_top,
            self.margin_right,
            self.margin_bottom,
            self.margin_left,
        ]


class HasPadding:
    empty = [None, None, None, None]

    def _init_padding(self, padding, top, right, bottom, left) -> None:

        if (padding, top, right, bottom, left) == (None, None, None, None, None):
            self.padding_top = None
            self.padding_right = None
            self.padding_bottom = None
            self.padding_left = None
            return

        if not isinstance(padding, list):
            padding = [padding]
        if len(padding) == 1:
            padding *= 4
        elif len(padding) == 2:
            padding *= 2
        elif len(padding) == 3:
            padding += [None]

        self.padding_top = padding[0]
        self.padding_right = padding[1]
        self.padding_bottom = padding[2]
        self.padding_left = padding[3]

        if top is not None:
            self.padding_top = top
        if right is not None:
            self.padding_right = right
        if bottom is not None:
            self.padding_bottom = bottom
        if left is not None:
            self.padding_left = left

    @property
    def padding(self) -> list[int | None]:
        return [
            self.padding_top,
            self.padding_right,
            self.padding_bottom,
            self.padding_left,
        ]

class HasBorder:
    """
    Child class is responsible for making sure each instance
    has variable, 'border_x' for each side.
    """

    empty = [None, None, None, None]
    negated = [False, False, False, False]
    default_weight = "thin"
    default_color = "000000"
    default = ("thin", "000000")
    valid_weights = (
        "dashDot",
        "dashDotDot",
        "dashed",
        "dotted",
        "double",
        "hair",
        "medium",
        "thick",
        "thin",
        "mediumDashDot",
        "mediumDashDotDot",
        "mediumDashed",
        "slantDashDot",
    )

    def _init_border(self, border, top, right, bottom, left) -> None:
        """
        Processes the full border and individual sides, where
        individual sides take priority only if they are not None
        """
        cls = type(self)
        self.border = border
        if top is not None:
            self.border_top = cls._interpret_single_value(top)
        if right is not None:
            self.border_right = cls._interpret_single_value(right)
        if bottom is not None:
            self.border_bottom = cls._interpret_single_value(bottom)
        if left is not None:
            self.border_left = cls._interpret_single_value(left)

        _ = self.border

    @property
    def border(self) -> list:
        """
        A great border
        """
        for side in ["border_top", "border_right", "border_bottom", "border_left"]:
            if not hasattr(self, side):
                setattr(self, side, None)

        cls = type(self)

        self.border_top = cls._interpret_single_value(self.border_top)
        self.border_right = cls._interpret_single_value(self.border_right)
        self.border_bottom = cls._interpret_single_value(self.border_bottom)
        self.border_left = cls._interpret_single_value(self.border_left)

        return [
            self.border_top,
            self.border_right,
            self.border_bottom,
            self.border_left,
        ]

    @border.setter
    def border(self, new: list) -> None:
        top, right, bottom, left = type(self)._parse_arg(new)
        self.border_top = top
        self.border_right = right
        self.border_bottom = bottom
        self.border_left = left

    @classmethod
    def _is_valid(cls, value: Any) -> bool:
        if value is None or value is False:
            return True
        if isinstance(value, tuple):
            if len(value) == 2:
                if isinstance(value[0], str) and isinstance(value[1], str):
                    if not value[1].startswith("#"):
                        return True
        return False

    @classmethod
    def _interpret_single_value(cls, value: Any) -> tuple[Any, Any]:
        """
        Given a value intended to represent a single border side, interpret
        it to one of the following valid formats:
        * None - unset, can be overriden
        * False - override parent and remove border
        * ('<weight>' | None | False, '<hex color>' | None | False)

        Valid inputs for ``value``:
        * None
        * True - converts to ``cls.default``
        * False
        * '<weight>' - we can tell if the string is in list of valid weights
        * '<hex color>'
        * ('<weight>' | None | True | False,)
        * ('<weight>' | None | True | False, '<hex color>' | None | True | False)
        """
        if cls._is_valid(value):
            return value

        # Treat 1-element tuple as single value
        if isinstance(value, tuple):
            if len(value) == 1:
                value = value[0]

        if value is True:
            return cls.default

        # If string, we can definitively conclude whether they were
        # referring to the weight or to the color.
        if isinstance(value, str):
            if value in cls.valid_weights:
                return (value, cls.default_color)
            else:
                return (cls.default_weight, value)

        # Now it must be a 2-element tuple
        if not isinstance(value, tuple):
            raise ValueError(f"Invalid border value, {value}")
        if not len(value) == 2:
            raise ValueError(f"Invalid border value, {value}")

        if value[0] is True:
            value = (cls.default_weight, value[1])

        if value[1] is True:
            value = (value[0], cls.default_color)

        for val in value:
            if not isinstance(val, str) and not val is None and not val is False:
                raise ValueError(f"Border weight/color values must be strings. {value} is invalid")

        if not value[0] in cls.valid_weights and not value[0] is None and not value[0] is False:
            raise ValueError(f"'{value[0]}' is not a valid weight")
        if not isinstance(value[1], str) and not value[1] is None and not value[1] is False:
            raise ValueError(f"'{value[1]}' is not a valid hex color")

        if isinstance(value[1], str):
            value = (value[0], value[1].lstrip("#"))
            if not len(value[1]) == 6:
                raise ValueError(f"Color value must be 6-character hex code. {value[1]} is invalid")

        return value

    @classmethod
    def _parse_arg(
        cls, border: bool | Iterable | None
    ) -> list:
        """
        Designed to mimic CSS border logic. Returns a 4-element list
        describing the border of 4 sides, in the order: top, right, bottom,
        left. Elements can either be None, False, or a string representing weight.

        Arguments of True will default to 'thin' border

        Example arguments to outputs:
            True or 'thin':
                [ 'thin', 'thin', 'thin', 'thin' ]
            [ 'thin', False ]:
                [ 'thin', False, 'thin', False ]
            [ 'thick', 'thick', 'thick' ]:
                [ 'thick', 'thick', 'thick', False ]
        """
        if border is None or border == cls.empty:
            return cls.empty
        if border == cls.negated:
            return cls.negated

        if isinstance(border, tuple):
            if len(border) > 2:
                print(border)
                raise TypeError("Border must be a list")

        if not isinstance(border, list):
            border = [border]

        for i, elem in enumerate(border):
            border[i] = cls._interpret_single_value(elem)

        if len(border) == 1:
            border = border * 4
        elif len(border) == 2:
            border += border
        elif len(border) == 3:
            border += [False]

        assert len(border) == 4, "Border must be 4 elements. If you're reading this, an excelbird developer made a mistake"
        return list(border)

    def _apply_border(self) -> None:
        if not hasattr(self, "__len__"):
            return
        if len(self) == 0 or self.border == [None, None, None, None]:
            return

        first = self[0]

        if len(self) == 1:
            if getattr(first, "is_empty", None) is True and hasattr(first, 'value'):
                first.value = ""
            first.border = self.border

        elif len(self) >= 2:
            mask = self._border_mask(*self.border)
            last = self[-1]
            middle_elements = self[1:-1]

            if getattr(first, "is_empty", None) is True and hasattr(first, 'value'):
                first.value = ""
            if getattr(last, "is_empty", None) is True and hasattr(last, 'value'):
                last.value = ""

            first.border = mask.first
            last.border = mask.last
            if len(self) > 2:
                for elem in middle_elements:
                    if getattr(elem, 'is_empty', None) is True and hasattr(elem, 'value'):
                        elem.value = ""
                    elem.border = mask.middle




from excelbird._layout_references import Globals

class HasId:
    """
    Has an id property which, when set, inserts a reference to
    self in `Globals.ids`.
    """

    @property
    def id(self):
        if not hasattr(self, "_id"):
            self._id = None
        return self._id

    @id.setter
    def id(self, new):
        self._set_id(new)

    def _set_id(self, new):
        if new is not None:
            if not isinstance(new, str):
                raise ValueError(f"Invalid id, `{new}`. Ids must be strings.")
            Globals.ids[new] = self
            if new.startswith("G::"):
                Globals.global_ids[new] = self
        self._id = new
        return self


class HasHeader:
    """
    Has an header property which, when set, inserts a reference to
    self in `Globals.headers`.
    """

    @property
    def header(self):
        if not hasattr(self, "_header"):
            self._header = None
        return self._header

    @header.setter
    def header(self, new):
        self._set_header(new)

    def _set_header(self, new):
        if new is not None:
            if not isinstance(new, str):
                raise ValueError(f"Invalid header, `{new}`. Headers must be strings.")
            Globals.headers[new] = self
            if new.startswith("G::"):
                Globals.global_headers[new] = self
        self._header = new
        return self



"""
xb.Schema
======

Where traditional dataframe schema classes define what should go *inside*
a dataframe's columns, `Schema` defines what those columns should be, 'when'
they should be, and how they got there. It defines a dataframe's 'state' at
a given point in time. And it provides the tools/methods needed to help you get to
that state, change it, and move to new states seamlessly.

It's a simple class, designed for ease of use and exceptional readability.

It's a subclass of dictionary.
    Keys: python-friendly variable names
    Values: namedtuple: ("Input col name", "Output col name (optional)")

Consider the following:

>>> sch_person = Schema(
...     first_name=("FName", "First Name"),
...     last_name=("LName", "Last Name"),
...     favorite_food=("Fav Food", "Favorite Food"),
... )
...
>>> sch_company = Schema(
...     comp_name=("Companyname", "Company Name"),
...     market_cap="Market Capitalization",
...     favorite_food="Preferred Employee Favorite Food",
... )
...
>>> sch_output = Schema(
...     sch_person[[
...         'last_name',
...         'age',
...     ]],
...     sch_company[[
...         'comp_name',
...     ]],
...     is_executive="Person is Executive"
... )

This should be readability-paradise. Given only the code above,
the reader should know exactly what's supposed to happen in the script:
    1. There are two input sources, person and company.
    2. The script will have to join person and company, and include fields from each.
    3. The script needs to add a new custom column, `is_executive`


Instance methods of this class try to provide as much utility as possible for
common operations. Here are just a few:
- `select_inputs()`: Take raw input data, validate our required columns can be
  found, select them, rename them python-friendly, and order them
- `select_outputs()`: Do the opposite, and again, validate that all the columns
  required by our output schema are present.
- `apply()`: Mid-workflow, safely re-order our columns and remove undesired ones.
"""
from __future__ import annotations

from pandas import DataFrame
from collections import namedtuple, ChainMap
from copy import copy
from excelbird.exceptions import SchemaError
from typing import overload


# The values held by each key of a Schema.
# Tuple's immutability helps enforce consistency in user's code
Column = namedtuple("Column", "input, output")


class Schema(dict):
    def __init__(
        self, *schemas, **kwargs: tuple[str, str] | tuple[str] | list[str] | str
    ) -> None:
        """Accepts keyword args only to enforce that var names are valid python"""
        if not all(isinstance(s, Schema) for s in schemas):
            raise TypeError("Positional args can only be existing Schemas")
        # Convert all passed values to tuples
        kwargs = {
            k: v if isinstance(v, tuple) else tuple(v) if isinstance(v, list) else (v,)
            for k, v in kwargs.items()
        }
        # For values of length 1 (input column only), infer the output column by duplicating it
        kwargs = {k: v if len(v) > 1 else v + v for k, v in kwargs.items()}

        if any(len(val) > 2 for val in kwargs.values()):
            raise ValueError(
                "Values must be a max of length 2. One input col name, and one output col name"
            )
        # Convert to Column
        kwargs = {k: Column(*v) for k, v in kwargs.items()}
        # If other schemas were passed in, create those as well
        # ChainMap combines dictionaries. We reverse the input first,
        # since for some reason ChainMap returns the values in opposite order
        super().__init__(**ChainMap(*tuple(reversed(schemas))), **kwargs)

    # def __getattr__(self, key: str) -> Column:
    #     """Lets you access dict items with dot notation"""
    #     if key in self.keys():
    #         return self[key]
    #     raise KeyError(f"Unknown key, '{key}'")

    @overload
    def __getitem__(self, key: list) -> Schema:
        ...

    @overload
    def __getitem__(self, key: str) -> Column:
        ...

    def __getitem__(self, key) -> Column | Schema:
        """
        Acts normal, unless you past a list.
        If a list is passed, filter and re-order the schema just like a
        dataframe. Returns new object
        """
        if not isinstance(key, list):
            return super().__getitem__(key)

        missing = [k for k in key if k not in self.keys()]
        if len(missing) > 0:
            raise SchemaError(f"Keys {missing} not present in schema")

        reordered = {copy(k): copy(self[k]) for k in key}
        return type(self)(**reordered)

    def __setitem__(self, key: str, val: Column | tuple[str, ...] | str) -> None:

        if isinstance(val, Column):
            return super().__setitem__(key, val)

        if isinstance(val, (list, tuple)):
            new = Column(*val)
            return super().__setitem__(key, new)

        if isinstance(val, str):
            new = Column(val, val)
            return super().__setitem__(key, new)

        raise ValueError(f"Invalid value, {val}")

    def drop(self, columns: list[str] | str) -> Schema:
        """
        Drop the specified keys
        """
        if not isinstance(columns, (list, tuple)):
            columns = [columns]

        return type(self)(
            **{copy(k): copy(v) for k, v in self.items() if k not in columns}
        )

    def apply(self, df: DataFrame, strict: bool = False) -> DataFrame:
        """
        - Filter ``df`` to remove columns that aren't in self
        - Re-order columns according to self's order

        If ``strict=True``, raise an error if ``df`` doesn't at least contain
        all columns specified in self.
        """
        if strict is False:
            return df[[k for k in self.keys() if k in df.columns]].copy()
        try:
            return df[[k for k in self.keys()]].copy()
        except KeyError:
            missing = [k for k in self.keys() if k not in df.columns]
            raise KeyError(
                f"Schema apply strict: The following columns were not found "
                f"in the dataframe (did you forget to run .select_inputs() first?): {missing}"
            )

    def rename(
        self,
        keys: dict | None = None,
        inputs: dict | None = None,
        outputs: dict | None = None,
    ) -> Schema:
        """
        Rename any part of self's items (key, input, output), by passing a dict
        who's keys are a current key in self, and values are the updated element.
        """
        new = self.copy()
        if keys is not None:
            new = type(self)()
            for key, val in self.items():
                if key not in keys:
                    new[key] = val
                else:
                    new[keys[key]] = val

        if inputs is not None:
            for key, new_name in inputs.items():
                new[key] = Column(new_name, new[key].output)

        if outputs is not None:
            for key, new_name in outputs.items():
                new[key] = Column(new[key].input, new_name)

        return new

    def update(self, new: dict | None = None, **kwargs) -> None:
        """
        Like ``dict.update()``, but if a regular ``dict`` or kwargs are passed,
        they're first used to create a new ``Schema`` before updating, so the
        correct format is maintained
        """
        if isinstance(new, type(self)):
            return super().update(new)
        if new is not None:
            return super().update(type(self)(**new))
        return super().update(type(self)(**kwargs))

    def rename_inputs_to_vars(self, df: DataFrame) -> DataFrame:
        """
        Take a dataframe in its input format and rename its columns to the
        python-friendly keys in self.
        """
        return df.rename(columns={val.input: key for key, val in self.items()})

    def rename_vars_to_outputs(self, df: DataFrame) -> DataFrame:
        """
        Rename columns to self's output names
        """
        return df.rename(columns={key: val.output for key, val in self.items()})

    def inputs(self) -> list[str]:
        return [val.input for val in self.values()]

    def outputs(self) -> list[str]:
        return [val.output for val in self.values()]

    def select_inputs(self, df: DataFrame) -> DataFrame:
        """
        Renames desired columns to var names, and selects them.
        If a column isn't found, an error is raised to force you to correct
        your schema.
        """
        missing = [col for col in self.inputs() if col not in df.columns]
        if len(missing) > 0:
            raise SchemaError(
                f"Schema requires input column(s), {missing}, not found in data."
            )
        df = self.rename_inputs_to_vars(df)
        return df[[k for k in self.keys()]]

    def select_outputs(self, df: DataFrame) -> DataFrame:
        """
        Renames var columns to their output names, and selects them.
        If any columns are missing, an error is raised to remind you to create them.
        """
        missing = [col for col in self.keys() if col not in df.columns]
        if len(missing) > 0:
            raise SchemaError(f"Please add columns, {missing} before outputting.")
        df = self.rename_vars_to_outputs(df)
        return df[[k for k in self.outputs()]]

    def reset_inputs(self) -> Schema:
        """
        Sets all inputs with output values. Use this if you're
        using a previous schema to read in data that was outputted from it
        """
        new = self.copy()
        for key in new.keys():
            new[key] = Column(new[key].output, new[key].output)
        return new

    def reset_outputs(self) -> Schema:
        """
        Fills all outputs with the current inputs
        """
        new = self.copy()
        for key in new.keys():
            new[key] = Column(new[key].input, new[key].input)
        return new

    def copy(self) -> Schema:
        return type(self)(**{copy(k): copy(v) for k, v in self.items()})

    def _repr_html_(self):

        return DataFrame(
            list(zip(self.inputs(), self.outputs())),
            columns=["Input", "Output"],
            index=list(self.keys()),
        )._repr_html_()



import re
from excelbird._formulae import FORMULAE

def main():

    func = """
    something + something 
    and then
    something else      is new
    here is ## not a comment

    here IS a comment,# some comment stuff
    another comment # stuff!
    And an ending comment # Thanks"""
    # print(format_string(func) + "X")
    s = "MAX(HOLIDAYS = 'O', h)"
    print(format_formula(s))



def format_string(s):
    if not isinstance(s, str):
        return s
    # Remove comments from end of string
    s = re.sub(r"#+.*?$", "", s)
    # Remove comments at the end of lines
    s = re.sub(r"#+.*?\n", " ", s)
    # Remove whitespace
    s = re.sub(r"\s+", " ", s)
    return s



def format_formula(s):
    s = re.sub(r"\([, ]+", r"(", s)  # comma before opening paren
    s = re.sub(r"[, ]+\)", r")", s) # trailing comma before closing paren
    s = re.sub(r"^= +", "=", s)  # leading space
    s = re.sub(r"[, ]+$", "", s)  # trailing comma
    s = re.sub(r"= *'(.*?)'", r'="\1"', s)  # using single quotes instead of double
    return s


def prefix_formulae_funcs(func: str) -> str:
    """
    Openpyxl will insert '@' before any function it doesn't think is a valid
    excel function. Its list of functions is outdated, so newer functions, like "CONCAT"
    will be interpreted as invalid incorrectly.

    Our solution is to find functions in the string, where there are capital letters, dots,
    or digits, followed by a starting parenthese, check if they are in openpyxl's
    `FORMULAE`, and if not, prefix them with "_xlfn."
    """

    matches = re.findall(r"([A-Z\.a-z0-9]+)\(", func)
    # matches = re.findall(r"^([A-Z\.0-9]+)\(", func) + re.findall(r"(?<![^\.A-Za-z0-9])([A-Za-z\.0-9]+)\(", func)
    matches = set([match for match in matches if match.upper() in FORMULAE])
    for match in matches:
        func = func.replace(match + "(", "_xlfn." + match + "(")
    return func



main()


    # s = """= _xlfn.MAX(, _xlfn.NETWORKDAYS(,,( _xlfn.MAX(B10, I13), J13, _xlfn.IF( _xlfn.OR(G13 = "O", G13="U-N"), 0, 'EY Holidays'!A2:'EY Holidays'!A31 ), ), 0 ) , """




"""
Detailed documentation and code examples coming soon.
"""
from __future__ import annotations
# External
import re
import pandas as pd
from typing import Any, overload
from copy import deepcopy
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl.styles import Font, PatternFill, Alignment, Border, Side
from openpyxl.styles.colors import Color

# Internal main
from excelbird._layout_references import Globals
from excelbird._base.identifier import HasId
from excelbird._base.styling import HasBorder
from excelbird._base.dotdict import Style
from excelbird._base.loc import Loc

from excelbird.styles import formats

from excelbird._utils.util import (
    get_dimensions,
)
from excelbird._utils.cell_util import (
    autofit_algorithm,
    remove_paren_enclosure,
    prefix_formulae_funcs,
    format_formula,
)
from excelbird._utils.color_algorithms import (
    color_is_light,
    get_alt_shade,
)
from excelbird.exceptions import AlreadyWrittenError, CellReferenceError
from excelbird._base.math import CanDoMath
from excelbird.core.expression import Expr
from excelbird.core.function import Func

cell_reference_warning_issued = False


class Cell(HasId, HasBorder, CanDoMath):
    """
    Represents a real cell in Excel. Capable of math operations in python, and tracking
    cell references on its own, without knowing locations.

    Excelbird cells are *not* mere containers for data. Since they represent a real object in Excel,
    each can only exist *once* inside a Book. This is because when you reference a Cell in a python
    expression (i.e. `cell_c = cell_a + cell_b`), cell references will be established so that when
    `cell_c` is written to the workbook, its value will be a formula that references the real
    locations of `cell_a` and `cell_b`. So if `cell_a` is found in multiple places in a Book,
    this formula could not be resolved.

    ALL attributes have a default of None. This allows parent containers to know whether a value
    has been set or not, to avoid override. For any attribute, if you just want to ignore the
    value set by a parent container, set its value to False, even if it isn't documented as accepting
    a boolean.

    Parameters
    ----------

    value : str or int or float, optional
        Value to display. If None, the Cell's styling will NOT be rendered. To display an empty
        cell with styling, pass an empty string as a value.
        This attribute will temporarily be None if a Cell references other Cells in an expression.
        For instance, if `cell3 = cell1 + cell2`, cell3's value will be None, until the parent
        book's .write() is called and cell1 and cell2's locations have been determined.
    dropdown : list or Cell or Col or Row, optional
        Apply data validation for the cell that offers a dropdown list of values to pick from. Pass
        a list of options, or a Cell, Col, or Row.
    id : str, optional
        Unique identifier to store globally so that this element can be referenced
        elsewhere in the layout without being assigned to a variable
    align_x : str, optional
        Horizontal alignment. Options: 'center', 'right', 'left'
    align_y : str, optional
        Vertical alignment. Options: 'top', 'center', 'bottom'
    indent : int or float, optional
        Indent the cell's value. Indentation direction will depend on horizontal alignment, so if
        align_x='right', the indentation will determine its distance from the right edge of the cell.
    center : bool, optional
        Center the cell horizontally and vertically. Shorthand for setting align_x and align_y to 'center'
    wrap : bool, optional
        Wrap the cell's text so that it doesn't continue outside of its boundary. As long as row_height=None
        and col_width=None, Excel will automatically resize the cell to display the full text.
    size : int, optional
        Font size
    bold : bool, optional
        Bold font
    italic : bool, optional
        Italic font
    color : str, optional
        Font color, as a hex code
    num_fmt : str, optional
        Cell value format. See the styles module for available formats
    currency : bool, optional
        Indicate that this Cell might contain a currency value. If value is an int or float
        and num_fmt has not been set, apply accounting format. The recommended use-case for this
        attribute is to set currency=True for a parent container once, and all of its numerical data
        will be formatted accordingly
    ignore_format : bool, optional
        Negate any number formatting set by a parent container. This is an easier alternative to
        setting num_fmt=False, currency=False
    fill_color : str, optional
        Hex code string color to fill the cell
    auto_color_font : bool, optional
        Sets font color to white or black, based on lightness of fill_color, so that text will be
        visible over any background. Lightness of fill_color is measured using the weighted euclidean
        norm of the rgb vector. If the resulting coefficient indicates a medium to light color, lightness
        will be re-calculated from the Luma of the rgb vector.
    auto_shade_font : bool, optional
        Font color will be a lighter or darker shade of fill_color. If fill_color is dark, a lighter shade
        will be chosen as the font color, and vice versa. Lightness coefficient of fill_color is measured
        using the weighted euclidean norm of the rgb vector. If the resulting coefficient indicates a
        medium to light color, lightness will be re-calculated from the Luma of the rgb vector.
    border : list[tuple or str or bool] or tuple[str or bool, str or bool] or str or bool, optional
        Syntax inspired by CSS. A non-list value will be applied to all 4 sides. If list,
        length can be 2, 3, or 4 elements. Order is [top, right, bottom, left]. If length 2,
        apply the first element to top and bottom border, and apply the second element to right and left.
    border_top : tuple[str or bool, str or bool] or str or bool, optional
        Top border. If True, a thin black border is used. If string (6 char hex code),
        use the default weight and apply the specified color. If string (valid weight name),
        use the default color and apply the specified weight. If tuple, apply the first
        element as weight, and second element as color.
    border_right : tuple[str or bool, str or bool] or str or bool, optional
        Right border. See border_top
    border_bottom : tuple[str or bool, str or bool] or str or bool, optional
        Bottom border. See border_top
    border_left : tuple[str or bool, str or bool] or str or bool, optional
        Left border. See border_top
    col_width : int, optional
        Column width. Format is the same as used in Excel.
    row_height : int, optional
        Row height. Format is the same as used in Excel.
    autofit : bool, optional
        Autofit column width based on the length of the value. This is NOT as accurate as the built-in
        autofit feature in Excel. This is because we can't determine the rendered width without actually
        rendering the value in the desired font and size, and counting pixels.
    merge : tuple[int, int], optional
        Merge this cell with other cells to its right or below. First element is the distance to
        merge below, and second element is the distance to merge across. For instance, `(0, 1)` will
        merge the current Cell with the one to the right. `(1,1)` will merge diagonally in a square.
        The cells being merged must have values of None. For instance, if you have a Row of multiple
        values and want the first and second elements to be merged, your code would be as follows:
        ``Row(Cell('a', merge=(0,1)), Cell(), Cell('b'), Cell('c'))`` - notice the second Cell is empty
    cell_style : dict, optional
        A dict that contains attributes to set. Priority is given to existing attributes - An attribute in
        cell_style will only be set if the Cell's attribute is currently None

    """

    _dimensions = 0
    elem_type = None

    @overload
    def __new__(cls, fn: str | set, **kwargs) -> Func:
        ...

    @overload
    def __new__(cls, func: str | set, **kwargs) -> Func:
        ...

    @overload
    def __new__(cls, ex: str | set, **kwargs) -> Expr:
        ...

    @overload
    def __new__(cls, expr: str | set, **kwargs) -> Expr:
        ...

    @overload
    def __new__(cls, mul: int, **kwargs) -> list[Cell | Func | Expr]:
        ...

    @overload
    def __new__(cls, *args, **kwargs) -> Cell:
        ...

    def __new__(cls, *args, fn=None, func=None, ex=None, expr=None, mul=None, **kwargs):
        fn = fn if fn is not None else func
        ex = ex if ex is not None else expr
        if mul is not None:
            elems = [cls.__new__(cls, *args, fn=fn, **kwargs) for _ in range(mul)]
            if len(elems) > 0:
                if isinstance(elems[0], cls):
                    for elem in elems:
                        elem.__init__(*args, **kwargs)
            return elems

        if fn is not None:
            new_func = Func.__new__(Func)
            new_func.__init__(fn, res_type=cls, **kwargs)
            return new_func

        if ex is not None:
            new_expr = Expr.__new__(Expr)
            new_expr.__init__(ex, res_type=cls, **kwargs)
            return new_expr

        return super().__new__(cls)

    def __init__(
        self,
        value: Any | None = None,
        dropdown: list | Any | None = None,  # list of values, Col, Row, or Cell
        id: str | None = None,
        align_x: str | None = None,
        align_y: str | None = None,
        indent: float | None = None,
        center: bool | None = None,  # center align horizontal and vertical
        wrap: bool | None = None,
        size: int | None = None,  # font size
        bold: bool | None = None,
        italic: bool | None = None,
        color: str | None = None,
        num_fmt: str | None = None,
        currency: bool | None = None,
        ignore_format: bool | None = None,
        fill_color: str | None = None,
        auto_color_font: bool | str | None = None,
        auto_shade_font: bool | float | None = None,
        border_left: bool | str | None = None,
        border_right: bool | str | None = None,
        border_top: bool | str | None = None,
        border_bottom: bool | str | None = None,
        border: bool | str | None = None,  # MUST be last border attr set
        col_width: int | None = None,  # finds the cell's column and adjust it
        row_height: int | None = None,  # finds the cell's row and adjusts it
        merge: tuple[int, int] | None = None,  # (y, x) counts of how many merges. Ex: (0, 1) merges cell with right adjacent
        _expr: list | None = None,
        _func: list | None = None,
        autofit: bool | None = None,
        cell_style: dict | None = None,
        _written: bool | None = None,
        fn: None = None,
        func: None = None,
        ex: None = None,
        expr: None = None,
        mul: None = None,
    ) -> None:
        del fn
        del func
        del ex
        del expr
        del mul
        self._written = False
        self._loc = None

        self.value = value
        self.dropdown = dropdown
        self.id = id
        self.align_x = align_x
        self.align_y = align_y
        self.indent = indent
        self.wrap = wrap
        self.size = size
        self.bold = bold
        self.italic = italic
        self.color = color
        self.num_fmt = num_fmt
        self.currency = currency
        self.ignore_format = ignore_format
        self.fill_color = fill_color
        self.auto_color_font = auto_color_font
        self.auto_shade_font = auto_shade_font
        self.col_width = col_width
        self.row_height = row_height
        self.merge = merge
        self._expr = _expr
        self._func = _func
        self.autofit = autofit
        self.center = center

        self._init_border(
            border,
            border_top,
            border_right,
            border_bottom,
            border_left,
        )
        self._inherit_style_without_override(cell_style)

        if isinstance(self.value, Cell):
            cell = self.value
            self.value = None
            new_dict = {
                k: v for k, v in cell.__dict__.items() if k not in ["_id", "_loc"]
            }
            for key, val in new_dict.items():
                if getattr(self, key) is None:
                    setattr(self, key, val)

        elif isinstance(self.value, (Expr, Func)):
            raise ValueError(
                "Can't pass Expr or Func as a Cell's `value`. "
                "Instead, use the expression/function by itself, and pass any "
                "Cell styling as a dict to `cell_style`."
            )

    @property
    def is_empty(self) -> bool:
        return self.value is None and self._func is None and self._expr is None

    @property
    def shape(self) -> tuple:
        return tuple()

    @property
    def width(self) -> int:
        return 1

    @property
    def height(self) -> int:
        return 1

    def ref(self, inherit_style: bool = False, **kwargs) -> Cell:
        """
        Get a new `Cell` who will display a cell reference to the current
        cell. This assumes that **both** the calling object
        and the returned object will be placed in the workbook.

        .. note::

            Calling ``.ref()`` is **not** necessary when an object is used in
            a python expression (i.e. ``some_cell + some_row``) and should `only`
            be used to duplicate data across a workbook.

        Parameters
        ----------
        inherit_style : bool, default False
            Copy the caller's style to the returned object.
        **kwargs : Any
            Extra keyword arguments are set as attributes on the returned
            object.

        Returns
        -------
        :class:`Cell <excelbird.Cell>`

        """
        if inherit_style is True:
            self_dict = deepcopy(self.__dict__)
            for key, val in self_dict.items():
                if key not in kwargs and key not in ["_id", "_loc", "_expr", "_func"]:
                    kwargs[key] = val
        return Cell(_expr=[self], **kwargs)

    def _expr_value(self) -> str | None:
        if self._expr is None:
            return None
        expr = str(self._eval_expr(self._expr))
        return remove_paren_enclosure(expr)

    def _func_value(self) -> str | None:
        if self._func is None:
            return None
        return self._eval_func(self._func)

    def _write(self) -> None:

        assert self._loc is not None, (
            "Excelbird developer error: Somehow, ._write() got called "
            "on a Cell that doesn't have a location. This is a serious issue!"
        )

        if self._written is True:
            raise AlreadyWrittenError(
                "Excelbird objects can only be written to a workbook once. This is "
                "because when `.write()` is called on a `Book`, the state of each of its elements "
                "changes: expressions are evaluated, series headers are inserted as actual cells, "
                "and cells with references are filled with the string locations of their references. "
                "\nSupport for repeated writes is possible, but hasn't been implemented yet."
            )

        if self._func is not None:
            self.value = "=" + self._func_value()
            self.value = self.value.replace(self._loc.title_str, "")
            self.value = prefix_formulae_funcs(self.value)
            self.value = format_formula(self.value)

        if self._expr is not None:
            self.value = self._expr_value()
            if self.value is None:
                return
            if "UNKNOWN" not in str(self.value):
                self.value = "=" + str(self.value)
            self.value = self.value.replace(self._loc.title_str, "")

        if self.value is None:
            return

        try:
            if pd.isnull(self.value):
                return
        except Exception:
            # Just making sure pd.isnull doesn't throw, if given a data type it doesn't like
            pass

        y, x = self._loc.y, self._loc.x
        cell = self._loc.cell
        cell.value = self.value

        # if ":" in str(self.value):
        #     self._loc.ws.formula_attributes['A5'] = {'t': 'array', 'ref': "A5:A5"}

        def get_dropdown() -> DataValidation | None:
            value = self.dropdown
            if value is None:
                return

            if type(value) in [list, tuple]:
                dropdown_items = [str(x) for x in value]
                formula = '"' + ",".join(dropdown_items) + '"'

            else:
                from excelbird.core.series import Col, Row

                if not isinstance(value, (Cell, Col, Row)):
                    raise ValueError(f"Invalid type for dropdown, {type(value)}")

                formula = None
                if isinstance(value, (Col, Row)):
                    formula = self._eval_expr(value.range()._expr).replace(
                        self._loc.title_str, ""
                    )
                else:
                    if value._loc is None:
                        raise ValueError(
                            "Cell reference in dropdown must be a valid Cell in workbook"
                        )

                    formula = self._eval_expr([value]).replace(self._loc.title_str, "")

            if formula is None:
                return None

            dv = DataValidation(type="list", formula1=formula, allow_blank=True)
            dv.add(cell)
            return dv

        validation = get_dropdown()
        if validation is not None:
            self._loc.ws.add_data_validation(validation)

        def get_number_format():
            if self.ignore_format is True:
                return

            if isinstance(self.num_fmt, str):
                return self.num_fmt

            if isinstance(self.value, str):
                return

            if isinstance(self.value, float):
                if self.currency is True:
                    return formats.number.accounting_float
                return formats.number.comma_float

            if isinstance(self.value, int):
                if self.currency is True:
                    return formats.number.accounting_int
                return formats.number.comma_int

        number_format = get_number_format()
        if number_format is not None:
            cell.number_format = number_format

        align, font, fill, border = {}, {}, {}, {}

        if self.center is True:
            align["horizontal"] = "center"
            align["vertical"] = "center"

        if self.align_x is not None:
            align["horizontal"] = self.align_x
        if self.align_y is not None:
            align["vertical"] = self.align_y

        if isinstance(self.indent, (int, float)):
            align["indent"] = self.indent

        if self.wrap is not None:
            align["wrap_text"] = self.wrap

        if self.size is not None:
            font["size"] = self.size
        if self.bold is not None:
            font["bold"] = self.bold
        if self.italic is not None:
            font["italic"] = self.italic
        if self.color is not None:
            font["color"] = self.color
        else:
            if self.auto_color_font is True and isinstance(self.fill_color, str):
                if not color_is_light(self.fill_color):
                    font["color"] = "FFFFFF"  # white
            elif self.auto_shade_font is not None and isinstance(self.fill_color, str):
                if isinstance(self.auto_shade_font, float):
                    font["color"] = get_alt_shade(self.fill_color, self.auto_shade_font)
                else:
                    font["color"] = get_alt_shade(self.fill_color)

        if self.fill_color is not None:
            fill = {"patternType": "solid", "fgColor": Color(self.fill_color)}

        def get_border(border) -> dict:
            def get_side(side) -> Side:
                if (
                    side is None
                    or side is False
                    or side == (None, None)
                    or side == (False, False)
                ):
                    return None
                assert isinstance(
                    side, tuple
                ), f"Internal developer error processing border. Border side value, {side} is invalid"
                if isinstance(side[0], str) and isinstance(side[1], str):
                    return Side(style=side[0], color=side[1])
                if isinstance(side[0], str):
                    return Side(style=side[0])
                if isinstance(side[1], str):
                    return Side(style=HasBorder.default_weight, color=side[1])
                raise ValueError(side)

            res = {}
            top, right, bottom, left = border
            if (side_top := get_side(top)) is not None:
                res["top"] = side_top
            if (side_right := get_side(right)) is not None:
                res["right"] = side_right
            if (side_bottom := get_side(bottom)) is not None:
                res["bottom"] = side_bottom
            if (side_left := get_side(left)) is not None:
                res["left"] = side_left
            return res

        border = get_border(self.border)

        if len(font) > 0:
            cell.font = Font(**font)
        if len(fill) > 0:
            cell.fill = PatternFill(**fill)
        if len(border) > 0:
            cell.border = Border(**border)
        if len(align) > 0:
            cell.alignment = Alignment(**align)

        if self.merge is not None:
            end_row = 1 + y + self.merge[0]
            end_column = 1 + x + self.merge[1]
            self._loc.ws.merge_cells(
                start_row=y + 1,
                start_column=x + 1,
                end_row=end_row,
                end_column=end_column,
            )

        if self.col_width is not None:
            self._loc.column_dimensions.width = self.col_width

        if self.autofit is True and self.col_width is None:
            curr = self._loc.column_dimensions.width
            new = autofit_algorithm(self.value)
            if new > curr:
                self._loc.column_dimensions.width = new

        if self.row_height is not None:
            self._loc.row_dimensions.height = self.row_height

        self._written = True

    def _set_loc(self, loc: Loc) -> None:
        self._loc = loc

    def __repr__(self):
        if self._expr is not None:
            return type(self).__name__ + "({...})"
        if self._func is not None:
            return type(self).__name__ + "(Func(...))"
        return f"{type(self).__name__}({self.value})"

    def _eval_func(self, func: list) -> str:
        def format_element(elem) -> str:
            if isinstance(elem, str):
                return elem

            if isinstance(elem, (int, float)):
                return str(elem)

            if get_dimensions(elem) > 0:
                cell_range = elem.range()._expr
                evaluated = self._eval_expr(cell_range)
                return remove_paren_enclosure(evaluated)

            if elem._loc is not None:
                return elem._loc.full_str

            if elem._expr is not None:
                evaluated = self._eval_expr(elem._expr)
                return remove_paren_enclosure(evaluated)

            if elem.value is not None:
                return str(elem.value)  # Don't put quotes around strings here

        res = "".join([format_element(e) for e in func])
        return res

    def _eval_expr(self, expr: list) -> str:
        def format_element(elem) -> str:
            if not isinstance(elem, Cell):
                return str(elem)

            if elem._loc is not None:
                return elem._loc.full_str

            if elem._expr is not None:
                return self._eval_expr(elem._expr)

            if elem._func is not None:
                return self._eval_func(elem._func)
            else:
                global cell_reference_warning_issued

                if elem.value is not None:
                    # if cell_reference_warning_issued is False:
                    #     print(
                    #         "Warning: A cell in your book is trying to reference a cell which "
                    #         "won't be placed in the book. The missing cell's value has been applied "
                    #         "as a hardcoded value in the valid cell's expression."
                    #     )
                    #     cell_reference_warning_issued = True

                    if isinstance(elem.value, str):
                        quote_stripped_val = elem.value.strip('"')
                        return f'"{quote_stripped_val}"'
                    else:
                        return str(elem.value)

                if (
                    cell_reference_warning_issued is False
                    and Globals.force_valid_references is False
                ):
                    CellReferenceError.issue_warning()
                    cell_reference_warning_issued = True
                else:
                    raise CellReferenceError()

                return "UNKNOWN"

        res = [format_element(e) for e in expr]

        if len(res) > 2:
            res = ["("] + res + [")"]

        return "".join(res)

    def _inherit_style_without_override(self, new_style: dict | Style | None) -> None:
        if new_style is not None:
            for key, val in new_style.items():
                check_unset = lambda x: x is None
                if key == "border":
                    check_unset = lambda x: x == [None, None, None, None]

                if check_unset(getattr(self, key, None)):
                    setattr(self, key, val)



"""
`Item` exists because most layout elements
(with the exception of :class:`Stack <excelbird.Stack>` / :class:`VStack <excelbird.VStack>` and :class:`Sheet <excelbird.Sheet>`)
can only hold **one** child type. In many cases, your code can be simplified, and
refactoring made easier, by using `Item` instead of the child type.

For instance, if you have a :class:`Frame <excelbird.Frame>` and want to refactor it
to a :class:`VFrame <excelbird.VFrame>`, you'd need to rename not only the container in question, but also
each of its children, changing all the :class:`Cols <excelbird.Col>` to :class:`Rows <excelbird.Row>`.
If `Item` were used instead of `Col`, this code change wouldn't be necessary, as each `Item`
would be interpreted as `Row` automatically by the new parent container type.

"""
from typing import Any

class Item:
    """
    Let the parent container decide what type the element should be.

    .. note::

        Once created, an `Item` cannot be modified, used in expressions/functions,
        or have its elements/attributes referenced, until it is either passed as an element
        to a parent container or resolved by calling :meth:`self.construct() <excelbird.Item.construct>`.

    Parameters
    ----------
    *args : Any
        All arguments are stored until the parent container decides which type should be
        instantiated, and will be passed to the corresponding type.
    **kwargs : Any
        All keyword arguments are stored until the parent container decides which type should be
        instantiated, and will be passed to the corresponding type.

    """

    def __init__(self, *args, **kwargs) -> None:
        self.__args = args
        self.__kwargs = kwargs

    def construct(self, dtype: type, **kwargs) -> Any:
        """
        Construct the desired type with the stored data.

        Parameters
        ----------
        dtype : type
            The desired type
        **kwargs : Any
            Additional keyword arguments will be passed to the constructor
            of the desired type

        Returns
        -------
        Any
            Type passed in argument ``dtype``
        """
        return dtype(*self.__args, **self.__kwargs, **kwargs)

    @classmethod
    def _resolve_all_in_container(cls, container: list, dtype: type):
        for i, elem in enumerate(container):
            if isinstance(elem, cls):
                container[i] = elem.construct(dtype)


class I(Item):
    """
    Shorthand for :class:`Item <excelbird.Item>`
    """

    pass





class SchemaError(Exception):
    pass

class AlreadyWrittenError(Exception):
    """
    An object can only be written once!
    """
    pass

class AutoOpenFileError(Exception):
    """
    For issues encountered when using `auto_open=True` in a Book.
    Common issues include:
        Not having Xlwings installed
        Trouble accessing file, usually caused by it being in OneDrive
    """
    pass

class InvalidSheetName(Exception):
    pass

class ExpressionResolutionError(Exception):
    default_msg = """
One or more expressions could not be resolved.
As a reminder, expression strings should reference elements inside
square brackets with no quotes, like: {"[some_col] + [other_col]"} or {"[1] + 99"}.

The value inside the brackets must contain one of the following:
    - A valid index of an element in the parent container
    - The `id` of any element defined globally
    - The `header` of any Col or Row defined globally

Another possible cause of this error is when setting `isolate=True` to a Sheet.
When True, the sheet will resolve its own references immediately after creation
and then clear the global memory of references, meaning any element created before,
or inside, an isolated sheet cannot be referenced in an Expr by id or header
after that sheet was created.
"""
    def __init__(self, message: str | None = None):
        if message is None:
            message = type(self).default_msg
        self.message = message
        super().__init__(self.message)


class ExpressionExecutionError(Exception):
    def __init__(self, original_exception: Exception, expr_string: str):
        super().__init__(
            f"An error was encountered when executing an Expr. This is usually a user syntax error. "
            "Excelbird tried to execute the following string as python code: '" + expr_string + "' "
            f"and encountered a '{type(original_exception).__name__}' with the description, '" + str(original_exception) + "'"
        )

class ExpressionTypeError(Exception):
    pass


class CellReferenceError(Exception):
    default_msg = """
One of the cells in your book is trying to reference a cell that isn't being
placed in the book. This would be fine if the missing cell contained references
to cells that eventually trace back to valid cells in the book, or if it had a value
of its own. In that case, the expression or value stored in the missing cell would be
inherited by those who were referencing it. The value of the missing cell is None.
Why was it referenced?
"""
    def __init__(self, message: str | None = None):
        if message is None:
            message = type(self).default_msg
        self.message = message
        super().__init__(self.message)
    
    @classmethod
    def issue_warning(cls, message: str | None = None):
        if message is None:
            message = cls.default_msg
        print(message)



"""
.. _header:

Headers
------------------------

A Series' :attr:`header <excelbird._Series.header>` attribute is a string which will
*later* be inserted as a :class:`Cell <excelbird.Cell>`. It has several purposes:

* Ignored by formulas and cell references, and is instead treated as a stylistic attribute
  of the parent container.
* Declared manually, or automatically if a :class:`pd.Series <pandas.Series>` is given
* Its styling is independent of the other child cells, and must be styled separately with :attr:`header_style <excelbird._Series.header_style>`.
* It can be referenced globally by other :class:`Exprs <excelbird.Expr>` in the layout, just like an ``id``. This is to avoid
  redundancy, and also provide convenience when headers are set automatically from :mod:`pandas` objects.

"""
# External
from __future__ import annotations
from pandas import Series
from numpy import ndarray
from typing import Iterable, Any, overload
from copy import copy, deepcopy

from excelbird._base.container import ListIndexableById
from excelbird._base.identifier import HasId
from excelbird._base.identifier import HasHeader
from excelbird._base.styling import HasBorder
from excelbird._base.dotdict import Style
from excelbird._base.loc import Loc
from excelbird._base.math import CanDoMath, elem_math

from excelbird._utils.util import (
    get_dimensions,
    get_idx,
    init_from_same_dimension_type,
)
from excelbird._utils.validation import (
    require_each_element_to_be_cls_type,
    ensure_value_is_not_number,
)
from excelbird._utils.argument_parsing import (
    combine_args_and_children_to_list,
    convert_all_to_type,
    move_remaining_kwargs_to_dict,
)

from excelbird.core.cell import Cell
from excelbird.core.expression import Expr
from excelbird.core.function import Func
from excelbird.core.gap import Gap
from excelbird.core.item import Item



class _Series(CanDoMath, ListIndexableById, HasId, HasHeader, HasBorder):
    _doc_primary_summary = """
    A 1-dimensional vector that can :ref:`be used in a python expression <expression_main>`,
    and can be prefixed with a :ref:`header <header>`.
    """
    _doc_params = """
    Parameters
    ----------
    *args : Union[Cell, Col, Row, str, int, float, list, tuple, pd.Series, np.ndarray, Gap, Item, Expr, Func, set]
        Children must be, or resolve to, `Cell`. Iterables will be *exploded* inplace. So, if given
        ``1, pandas.Series([2, 3]), 4``, it will be read as ``1, 2, 3, 4`` before converting the
        ints to `Cell`. If a :class:`pandas.Series` is passed as the *only* argument, its name
        (if present) will be used as :ref:`header <header>`.
    children : list, optional
        Will be combined with args
    id : str, optional
        Unique identifier to store globally so that this element can be referenced
        elsewhere in the layout without being assigned to a variable
    header : str, optional
        Unique identifier to be inserted as a Cell at position 0 at write time. Can be referenced
        by an `Expr` elsewhere in the layout. Set automatically if a named ``pandas.Series`` is given.
        :ref:`Read more <header>`
    sep : Gap or bool or int or dict, optional
        A sep in any excelbird layout element inserts a `Gap` between each of its children.
        If True, a default of ``Gap(1)`` is used. If int, ``Gap(sep)`` will be used. If a dict,
        ``Gap(1, **sep)`` will be used.
    background_color : str, optional
        Hex code for background color. Will be applied to any Gap child who hasn't specified its own
        fill_color.
    cell_style : dict, optional
        Each key/value will be used to set an attribute on each child Cell (header excluded)
        only if the respective attribute has not already been set on the child Cell (its value is None).
        This mimics HTML/CSS behavior, where styling declared at the parent level is passed down
        to children, but each child can override the parent.
    header_style : dict, optional
        Just like cell_style, but for the header only. Ignored if header is None.
    border : list[tuple or str or bool] or tuple[str or bool, str or bool] or str or bool, optional
        Syntax inspired by CSS. A non-list value will be applied to all 4 sides. If list,
        length can be 2, 3, or 4 elements. Order is [top, right, bottom, left]. If length 2,
        apply the first element to top and bottom border, and apply the second element to right and left.
        To apply border to children instead, use cell_style.
    border_top : tuple[str or bool, str or bool] or str or bool, optional
        Top border. If True, a thin black border is used. If string (6 char hex code),
        use the default weight and apply the specified color. If string (valid weight name),
        use the default color and apply the specified weight. If tuple, apply the first
        element as weight, and second element as color.
    border_right : tuple[str or bool, str or bool] or str or bool, optional
        Right border. See border_top
    border_bottom : tuple[str or bool, str or bool] or str or bool, optional
        Bottom border. See border_top
    border_left : tuple[str or bool, str or bool] or str or bool, optional
        Left border. See border_top
    **kwargs : Any
        Remaining kwargs will be applied to cell_style
    """

    _dimensions = 1
    elem_type = Cell

    @overload
    def __new__(cls, fn: str | set, **kwargs) -> Func:
        ...

    @overload
    def __new__(cls, func: str | set, **kwargs) -> Func:
        ...

    @overload
    def __new__(cls, ex: str | set, **kwargs) -> Expr:
        ...

    @overload
    def __new__(cls, expr: str | set, **kwargs) -> Expr:
        ...

    @overload
    def __new__(cls, *args, **kwargs) -> _Series:
        ...

    def __new__(cls, *args, fn=None, func=None, ex=None, expr=None, **kwargs):
        fn = fn if fn is not None else func
        ex = ex if ex is not None else expr
        if fn is not None:
            new_func = Func.__new__(Func)
            new_func.__init__(fn, res_type=cls, **kwargs)
            return new_func

        if ex is not None:
            new_expr = Expr.__new__(Expr)
            new_expr.__init__(ex, res_type=cls, **kwargs)
            return new_expr

        return super().__new__(cls)

    def __init__(
        self,
        *args: Any,
        children: list | None = None,
        id: str | None = None,
        header: str | None = None,
        sep: Any | None = None,
        border_left: bool | str | None = None,
        border_right: bool | str | None = None,
        border_top: bool | str | None = None,
        border_bottom: bool | str | None = None,
        border: bool | str | Iterable | None = None,
        background_color: str | None = None,
        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        fn: None = None,
        func: None = None,
        ex: None = None,
        expr: None = None,
        **kwargs,
    ) -> None:
        del fn
        del func
        del ex
        del expr
        children = combine_args_and_children_to_list(args, children)

        children = [i for i in children if i is not None]

        if cell_style is None:
            cell_style = dict()
        if header_style is None:
            header_style = dict()

        if len(children) == 1 and isinstance(get_idx(children, 0), Series):
            if children[0].name is not None and header is None:
                header = children[0].name

        self._format_args(children)

        move_remaining_kwargs_to_dict(kwargs, cell_style)

        self._loc = None
        self.id = id
        self.header = header
        self.background_color = background_color
        self.header_style = Style(**header_style)
        # Dicts that must be passed to children
        self.cell_style = Style(**cell_style)

        self._init(children)

        self._init_border(
            border,
            border_top,
            border_right,
            border_bottom,
            border_left,
        )
        if sep is not None:
            self._insert_separator(sep)

        self.header_written = False

    def from_valid_children(
        self,
        children: list,
        id: str | None = None,
        header: str | None = None,
        sep: Any | None = None,
        border_left: bool | str | None = None,
        border_right: bool | str | None = None,
        border_top: bool | str | None = None,
        border_bottom: bool | str | None = None,
        border: bool | str | Iterable | None = None,
        background_color: str | None = None,
        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        **kwargs,
    ):
        if cell_style is None:
            cell_style = dict()
        if header_style is None:
            header_style = dict()

        move_remaining_kwargs_to_dict(kwargs, cell_style)

        self._loc = None
        self.id = id
        self.header = header
        self.background_color = background_color
        self.header_style = Style(**header_style)
        # Dicts that must be passed to children
        self.cell_style = Style(**cell_style)

        self._init(children)

        self._init_border(
            border,
            border_top,
            border_right,
            border_bottom,
            border_left,
        )
        if sep is not None:
            self._insert_separator(sep)

        self.header_written = False

    def ref(self, inherit_style: bool = False, **kwargs):
        """
        Get a new object with cell references to those in the caller.
        This assumes that **both** the calling object
        and the returned object will be placed in the workbook.

        .. note::

            Calling ``.ref()`` is **not** necessary  when an object is used in
            a python expression (i.e. ``some_cell + some_row``) and should `only`
            be used to duplicate data across a workbook.

        Parameters
        ----------
        inherit_style : bool, default False
            Copy the caller's style to the returned object.
        **kwargs : Any
            Extra keyword arguments are set as attributes on the returned
            object.

        Returns
        -------
        :class:`Self`

        Notes
        -----

        .. note::

            ``self.header`` is a stylistic attribute, and therefore will **not** be
            passed to the returned object unless ``inherit_style=True``. And, if style
            is inherited, the header will be copied to its new cell, instead of having
            a cell reference made to it.

        """
        new_elements = [
            i.ref(inherit_style=inherit_style, **kwargs)
            for i in self
        ]
        new_dict = kwargs
        if inherit_style is True:
            self_dict = deepcopy(self.__dict__)
            for key, val in self_dict.items():
                if key == "_header":
                    key = "header"
                if key not in new_dict and key not in ["_id", "_loc"]:
                    new_dict[key] = val
        return type(self)(*new_elements, **new_dict)

    def transpose(self, **kwargs):
        """
        Convert to sibling type. Places current children into the returned object,
        without copying or making cell references to them.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to apply as attributes to the new object.

        Returns
        -------
        :class:`Col <excelbird.Col>` or :class:`Row <excelbird.Row>`
            The opposite to self's type. Try ``type(my_obj).sibling_type``

        Notes
        -----
        **Assumes that the caller won't be placed in the layout**. Do not
        place both the calling object and returned object in the layout, since
        they both contain the same children.

        .. code-block::

            # 'current' must not be placed in the workbook.
            new = current.transpose()

        To include the caller and make cell references to it, get a reference
        first:

        .. code-block::

            new = current.ref().transpose()

        """
        elements = list(self)
        new = type(self).sibling_type(*elements)
        for key, val in self.__dict__.items():
            if key == "_header":
                key = "header"
            if key == "_id":
                key = "id"
            setattr(new, key, val)

        for key, val in kwargs.items():
            if hasattr(new, key):
                setattr(new, key, val)
            elif hasattr(new, 'cell_style'):
                new.cell_style[key] = val
        return new

    def range(self, include_headers: bool = False) -> Cell:
        """
        Get a reference to the range of the series, instead of a vector of
        cell references.

        Parameters
        ----------
        include_headers : bool, default False
            If True, the header cell will be included in the range reference.

        Returns
        -------
        :class:`Cell <excelbird.Cell>`
        """
        if self.header_written is True and include_headers is False:
            first = self[1]
        else:
            first = self[0]
        last = self[-1]
        return first >> last

    @property
    def shape(self) -> tuple[int]:
        length = sum([1 if not isinstance(i, Gap) else i for i in self])
        if self.header is not None:
            length += 1
        return (length,)

    def _format_args(self, args: list) -> None:
        self._explode_all_1d_iterables(args)
        convert_all_to_type(args, set, Expr)
        convert_all_to_type(args, int, Cell, strict=True)
        convert_all_to_type(args, (str, float), Cell)
        Item._resolve_all_in_container(args, type(self).elem_type)
        for i, elem in enumerate(args):
            if not isinstance(elem, (Cell, Iterable, Gap, Expr, Func)):
                args[i] = Cell(elem)

    def _explode_all_1d_iterables(self, args: list) -> None:

        for i, elem in enumerate(args):
            if isinstance(elem, Series):
                sr = args.pop(i)
                for cell in reversed(sr.reset_index(drop=True)):
                    args.insert(i, cell)

            elif isinstance(elem, (list, tuple, ndarray)):
                sr = args.pop(i)
                for value in reversed(sr):
                    args.insert(i, value)

    def _resolve_background_color(self) -> None:
        for elem in self:
            if hasattr(elem, "_resolve_background_color"):
                if (
                    self.background_color not in [None, False]
                    and elem.background_color is None
                ):
                    elem.background_color = self.background_color
                elem._resolve_background_color()

        if self.background_color not in [None, False]:
            for elem in self:
                if isinstance(elem, Gap):
                    if "fill_color" not in elem.kwargs:
                        elem.fill = True
                        elem.kwargs["fill_color"] = self.background_color

    def _resolve_gaps(self):
        Gap._explode_all_to_values(self, Cell)

    def _set_loc(self, loc: Loc) -> None:
        self._loc = loc

        offset = self._starting_offset()
        for elem in self:
            elem._set_loc(
                Loc((self._loc.y + offset.y, self._loc.x + offset.x), self._loc.ws)
            )
            offset = self._inc_offset(offset, elem)

    def __getitem__(self, key):
        if not isinstance(key, list):
            # return super().__getitem__(key)
            return ListIndexableById.__getitem__(self, key)

        new_elements = [self[self._key_to_idx(k)] for k in key]
        new_dict = {k: v for k, v in self.__dict__.items() if k not in ["_id", "_loc"]}
        if "_header" in new_dict:
            new_dict["header"] = new_dict.pop("_header")

        return type(self)(*new_elements, **new_dict)

    # def __rshift__(self, other):
    #     if get_dimensions(other) < get_dimensions(self):
    #         return elem_math(self[0], other, lambda a, b: a >> b, " >> ")
    #     return self[0] >> other[-1]
    #
    # def __rrshift__(self, other):
    #     if get_dimensions(other) < get_dimensions(self):
    #         return elem_math(other, self[-1], lambda a, b: a >> b, " >> ")
    #     return other[0] >> self[-1]

    def _validate_child_types(self) -> None:
        cls_name = type(self).__name__
        elem_type_name = type(self).elem_type.__name__
        valid_types = (
            type(self).elem_type,
            Gap,
        )
        for elem in self:
            if not isinstance(elem, valid_types):
                raise TypeError(
                    f"At write time, a {cls_name} can only hold {elem_type_name}s or Gaps. "
                    "To arrange mixed types, place them in a Stack or VStack"
                )
            if hasattr(elem, "_validate_child_types"):
                elem._validate_child_types()

    def _write(self) -> None:
        require_each_element_to_be_cls_type(self)

        self._apply_border()

        for cell in self:
            cell._inherit_style_without_override(self.cell_style)

        if self.header is not None:
            ensure_value_is_not_number(self.header)
            new_header = Cell(self.header)

            new_header._set_loc(self._loc)

            new_header._inherit_style_without_override(self.header_style)

            if (
                self.cell_style.get("autofit") is True
                and self.header_style.get("autofit") is not False
            ):
                new_header.autofit = True

            self.insert(0, new_header)
            self.header_written = True

        for cell in self:
            cell._write()

    def _starting_offset(self) -> Loc:
        ...


class Row(_Series):
    ...


class Col(_Series):
    _doc_custom_summary = """
    * Direction: **vertical**
    * Child Type: :class:`Cell`
    """

    sibling_type: type = Row  # these are set after class declaration
    elem_type = Cell

    def transpose(self, **kwargs) -> Row:
        return super().transpose(**kwargs)

    def ref(self, inherit_style: bool = False, **kwargs) -> Col:
        return super().ref(inherit_style, **kwargs)

    @property
    def width(self):
        return 1

    @property
    def height(self):
        return self.shape[0]

    def _repr_html_(self):
        from pandas import DataFrame

        header = "" if self.header is None else self.header
        elems_to_show = list(self) if len(self) <= 10 else list(self)[:10]

        if len(self) > 10:
            elems_to_show.append(f"(+{len(self)-10})")

        df = DataFrame(elems_to_show + [""], columns=[header])
        return df.style.hide(axis="index")._repr_html_()

    def _border_mask(self, top, right, bottom, left) -> Style:
        return Style(
            first=[top, right, False, left],
            last=[False, right, bottom, left],
            middle=[False, right, False, left],
        )

    @staticmethod
    def _inc_offset(offset: Loc, elem: Any) -> Loc:
        offset.y += elem.height
        return offset

    def _starting_offset(self) -> Loc:
        offset = Loc((0, 0), self._loc.ws)
        if getattr(self, "_header", None) is not None:
            offset.y += 1
        return offset


class Row(_Series):
    _doc_custom_summary = """
    * Direction: **horizontal**
    * Child Type: :class:`Cell`
    """
    sibling_type = Col  # these are set after class declaration
    elem_type = Cell

    def transpose(self, **kwargs) -> Col:
        return super().transpose(**kwargs)

    def ref(self, inherit_style: bool = False, **kwargs) -> Row:
        return super().ref(inherit_style, **kwargs)

    @property
    def width(self):
        return self.shape[0]

    @property
    def height(self):
        return 1

    def _repr_html_(self):
        from pandas import DataFrame

        header = "" if self.header is None else self.header
        elems_to_show = list(self) if len(self) <= 10 else list(self)[:10]

        if len(self) > 10:
            elems_to_show.append(f"(+{len(self)-10})")

        df = DataFrame(
            [elems_to_show, ["" for _ in elems_to_show]],
            index=[header, ""],
            columns=["" for _ in elems_to_show],
        )

        if self.header is None:
            return df.style.hide(axis="index")._repr_html_()
        return df._repr_html_()

    def _border_mask(self, top, right, bottom, left) -> Style:
        return Style(
            first=[top, False, bottom, left],
            last=[top, right, bottom, False],
            middle=[top, False, bottom, False],
        )

    @staticmethod
    def _inc_offset(offset: Loc, elem: Any) -> Loc:
        offset.x += elem.width
        return offset

    def _starting_offset(self) -> Loc:
        offset = Loc((0, 0), self._loc.ws)
        if getattr(self, "_header", None) is not None:
            offset.x += 1
        return offset


Col.__doc__ = f"""
    {_Series._doc_primary_summary}

    {Col._doc_custom_summary}

    {_Series._doc_params}

    """

Row.__doc__ = f"""
    {_Series._doc_primary_summary}

    {Row._doc_custom_summary}

    {_Series._doc_params}

    """







# External
from typing import Any

# Internal main
from excelbird._layout_references import Globals
from excelbird._base.dotdict import Style
from excelbird.styles.styles import default_table_style
from excelbird._utils.util import (
    get_idx,
)
from excelbird._utils.argument_parsing import (
    combine_args_and_children_to_list,
    move_remaining_kwargs_to_dict,
)
from excelbird._utils.pass_attributes import (
    pass_dict_to_children,
)
# Internal core
from excelbird.core.gap import Gap
from excelbird.core.cell import Cell
from excelbird.core.series import Col, Row
from excelbird.core.frame import Frame, VFrame
from excelbird.core.stack import VStack, Stack
from excelbird.core.expression import Expr
from excelbird.core.function import Func


class Sheet(VStack):
    """
    Behaves similar to :class:`VStack` - it can hold any element, and arranges its children vertically - but
    lacks some styling features like margin and padding.

    * Direction: **vertical**
    * Child Type: :class:`Stack`, :class:`VStack`, :class:`Frame`, :class:`VFrame`, :class:`Col`, :class:`Row`, :class:`Cell`

    .. note:: If the first argument in ``*args`` is a string, it will be used as the ``title`` attribute. This allows for better readability in complex layouts with multiple sheets, so a sheet's title can be visible at the start of the container.

    Parameters
    ----------
    *args : Union[Stack, VStack, Frame, VFrame, Col, Row, Cell, list, tuple, str, int, float, pd.Series, pd.DataFrame, np.ndarray, Gap, Expr, Func, set]
        Can take any layout element (besides `Book` and `Sheet`) or any value that can
        be used to construct a layout element.
    children : list, optional
        Will be combined with args
    title : str, optional
        Sheet name
    sep : Gap or bool or int or dict, optional
        A sep in any excelbird layout element inserts a Gap between each of its children.
        If True, a default of Gap(1) is used. If int, Gap(sep) will be used. If a dict,
        ``Gap(1, **sep)`` will be used.
    tab_color : str, optional
        Hex color for tab color.
    end_gap : bool or int or dict or Gap, optional
        Applies a Gap to cells below and to the right of the Sheet. The Gap determines
        the number of columns filled, and 1/3 the number of rows filled. The default
        is Gap(35, fill_color="FFFFFF") (white). This means apply whitespace (hide grid)
        for 35 columns, and 105 rows surrounding the Sheet contents.
    isolate : bool, optional
        After initialization, clear the global memory of ids and headers, so references
        in future declared Sheets won't conflict with previous ones. This will also isolate
        previously declared Sheets, so they musn't reference elements declared after the current
        one.
    hidden : bool, optional
        Whether to hide the Sheet
    zoom : int, optional
        Percentage zoom level. (Passing None or 100 will have the same effect)
    background_color : str, optional
        Hex code for background_color. Will be applied to fill_color of any Gap child who hasn't
        specified its own fill_color, and to any child Stack/VStack's margins. Will also be passed
        down to any child (Cell excluded) who hasn't specified its own background_color.
    cell_style : dict, optional
        Applied to each child who has cell_style
    header_style : dict, optional
        Applied to each child who has header_style
    table_style : dict or bool, optional
        Applied to each child who has table_style
    **kwargs : Any
        Remaining kwargs will be applied to self.cell_style

    """

    _dimensions = -1

    def __init__(
        self,
        *args: Any,
        children: list | None = None,
        title: str | None = None,
        sep: Any | None = None,
        tab_color: str | None = None,
        end_gap: bool | int | dict | Gap | None = None,
        isolate: bool | None = None,
        hidden: bool | None = None,
        zoom: int | None = None,
        background_color: str | None = None,
        cell_style: Style | dict | None = None,
        header_style: Style | dict | None = None,
        table_style: Style | dict | bool | None = None,
        **kwargs,
    ) -> None:
        children = combine_args_and_children_to_list(args, children)
        first_arg = get_idx(children, 0)
        if isinstance(first_arg, str):
            title = children.pop(0)

        children = [i for i in children if i is not None]

        # Alternative to init_from_same_dimension_type
        if len(children) == 1 and isinstance(first_arg, Sheet):
            children = list(first_arg)
            new_kwargs = first_arg.__dict__
            new_kwargs.pop("_loc")
            for key, val in new_kwargs.items():
                setattr(self, key, val)

        if cell_style is None:
            cell_style = dict()
        if header_style is None:
            header_style = dict()
        if table_style is None or table_style is False:
            table_style = dict()
        elif table_style is True:
            table_style = default_table_style

        self._format_args(children)

        move_remaining_kwargs_to_dict(kwargs, cell_style)

        self._loc = None
        self.title = title
        self.tab_color = tab_color
        self.end_gap = end_gap
        self.isolate = isolate
        self.hidden = hidden
        self.zoom = zoom
        self.background_color = background_color
        # Dicts that must be passed to children
        self.cell_style = Style(**cell_style)
        self.header_style = Style(**header_style)
        self.table_style = Style(**table_style)

        self._init(children)

        if sep is not None:
            self._insert_separator(sep)

        if self.isolate is True:
            self._resolve_all_references()
            self._resolve_all_references()
            self._resolve_all_references()
            #     raise ExpressionResolutionError(
            #         "Couldn't resolve all expression references.\nThis error was raised during the "
            #         f"creation of sheet, '{self.title}'.\nWhen `isolate=True` for "
            #         "a sheet, it will try to resolve all of its expression references "
            #         "immediately after being created, and then clear the global memory of references "
            #         "so any element created afterwards won't be able to reference them."
            #     )
            Globals.clear_references()

    def _resolve_all_references(self) -> bool:
        Expr._set_use_ref_for_container_recursive(self)

        all_resolved = False
        attempts = 0
        while not all_resolved and attempts <= 5:
            all_resolved = True
            attempts += 1
            if Expr._resolve_container_recursive(self) is False:
                all_resolved = False
            if Func._resolve_container_recursive(self) is False:
                all_resolved = False

        return all_resolved

    def _apply_end_gap(self) -> None:
        gap = self.end_gap
        if gap is None or gap is False:
            return

        if (
            not type(gap) in [bool, int]
            and not isinstance(gap, Gap)
            and not isinstance(gap, dict)
        ):
            raise ValueError("end_gap must be bool, int, Gap, dict")

        default_size = 35
        # default_color = "FFFFFF"  # white

        if gap is True:
            gap = Gap(default_size)
        elif type(gap) == int:
            gap = Gap(gap)
        elif isinstance(gap, dict):
            if "size" in gap:
                gap = Gap(gap.pop("size"), **gap)
            else:
                gap = Gap(default_size, **gap)

        if "fill_color" not in gap.kwargs and self.background_color is not None:
            gap.kwargs["fill_color"] = self.background_color

        if "row_multiplier" not in gap.kwargs:
            gap.kwargs["row_multiplier"] = 3

        size, kwargs = int(gap), gap.kwargs
        row_multiplier = kwargs.pop("row_multiplier")
        width, height = self.width, self.height
        new_elements = []
        for i, elem in reversed(list(enumerate(self))):
            new_elements.insert(0, self.pop(i))
        new_cols = size
        new_rows = size * row_multiplier
        full_height = height + new_rows
        self.append(
            Stack(
                VStack(
                    *new_elements,
                    VFrame(
                        *[
                            Row(*[Cell("", **kwargs) for _ in range(width)])
                            for _ in range(new_rows)
                        ]
                    ),
                ),
                Frame(
                    *[
                        Col(*[Cell("", **kwargs) for _ in range(full_height)])
                        for _ in range(new_cols)
                    ]
                ),
            )
        )

    def _resolve_gaps(self) -> None:
        super()._resolve_gaps()
        self._apply_end_gap()

    def _write(self) -> None:

        if self.tab_color is not None:
            self._loc.ws.sheet_properties.tabColor = self.tab_color

        if self.hidden is True:
            self._loc.ws.sheet_state = "hidden"

        if self.zoom is not None:
            self._loc.ws.sheet_view.zoomScale = self.zoom

        pass_dict_to_children(self, "cell_style")
        pass_dict_to_children(self, "header_style")
        pass_dict_to_children(self, "table_style")

        for elem in self:
            elem._write()

        if self.isolate is True:
            Globals.clear_references(self._loc.ws.title)



import webbrowser
import click
import os
import shutil

@click.group()
def main():
    pass

@click.command()
@click.argument('action', required=False)
@click.option('--how', '-h', help="What command to use for the build", type=click.Choice(['make', 'sphinx-build', 'fast']), default='make', show_default=True)
def docs(action, how):
    here = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(here, os.pardir, 'docs')

    if action is None:
        docs_index = os.path.join(path, 'build', 'html', 'index.html')
        webbrowser.open_new_tab('file://' + docs_index)

    elif action == "build":
        base_path = os.path.join(path, os.pardir, os.pardir)

        if how == 'fast':
            os.system(f"cd {base_path}; source env/bin/activate; cd excelbird/docs; make html")
            return

        shutil.rmtree(os.path.join(path, 'build'))
        if how == 'make':
            os.system(f"cd {base_path}; source env/bin/activate; cd excelbird/docs; make html")
        elif how == 'sphinx-build':
            os.system(f"cd {base_path}; source env/bin/activate; cd excelbird/docs; sphinx-build -b html source build/html")




# @click.command()
# @click.argument('base')
# @click.argument('action')
# def temp(base, action=None):
#     if base == "docs":
#         here = os.path.abspath(os.path.dirname(__file__))
#         docs_html = os.path.join(here, os.pardir, 'docs', 'build', 'html')
#         index_html = os.path.join(docs_html, 'index.html')
#         webbrowser.open_new_tab('file://' + index_html)


main.add_command(docs)

if __name__ == "__main__":
    main()
